#pragma once

#include <glew.h>
#include <QtGui/QtGui>
#include <QGLWidget>
#include <gl.h>
#include <glu.h>

#include <framebufferObject.h>
#include <texture2D.h>
#include <gpuProgram.h>
#include <gpuShader.h>

//#include "vectorization.h"
#include "tracking.h"
#include "skeleton.h"

using namespace gpu;

enum InteractionMode { MODE_NONE=0, MODE_MOVE, MODE_DELETE, MODE_SPLIT, MODE_DRAW};

class ImagePaint : public QGLWidget {
	Q_OBJECT
		Q_PROPERTY(QImage image READ image WRITE setImage NOTIFY imageChanged(const QImage&)) 
		Q_PROPERTY(double zoom READ zoom WRITE setZoom NOTIFY zoomChanged(double)) 
		Q_PROPERTY(double resolution READ resolution WRITE setResolution NOTIFY resolutionChanged)
		Q_PROPERTY(double originX READ originX WRITE setOriginX NOTIFY originXChanged)
		Q_PROPERTY(double originY READ originY WRITE setOriginY NOTIFY originYChanged)
public:
	ImagePaint(QWidget *parent);
	~ImagePaint();

	virtual void restoreSettings(QSettings& settings);
	virtual void saveSettings(QSettings& settings);

	QSize imageSize() const { return image().size(); }
	const QImage& image() const { return m_image; }
	QPointF origin() const { return m_moveOrigin; }
	double originX() const { return m_moveOrigin.x(); }
	double originY() const { return m_moveOrigin.y(); }
	double zoom() const { return m_zoom; }
	double resolution() const { return m_resolution; }
	double scale() const { return m_zoom * m_resolution; }

	virtual QPointF view2image(const QPointF& p) const;
	virtual QPointF image2zoom(const QPointF& p) const;
	virtual QPointF image2view(const QPointF& p) const;
	virtual QTransform viewTransform(const QSizeF& sz, double zoom, const QPointF& origin) const;
	virtual float pt2px(float pt) const;

	void setOverlay(QWidget *overlay);
	void setInteractionMode(InteractionMode mo);

public slots:
		void setImage(const QImage& image);
		void setOriginX(double value);
		void setOriginY(double value);
		void setZoom(double value);
		void setResolution(double value);
		void zoomIn();
		void zoomOut();
		void reset();
		void hold();
		void toggle();
		void copy();
		void savePNG(const QString& text=QString());
		void saveSnapshot(QString &text=QString());
		void loadSkeletonFile();
		void saveSkeletonFile();
		void loadSVGFile();
		void saveSVGFile();
		void loadParameterFile();
		void saveParameterFile();
		void saveEverythingFile();

		void reloadShaders();
		void onIndexChanged(int);
		void refresh() { update(); }

		void nextFrame();
		void ComputeSkeletonChicked(bool isChecked);
		void setStreamline(bool);
		void setDrawArrowField(bool);
		void setStreamlineType(int);
		void branchPushButtonClicked();

signals:
		void imageChanged(const QImage&);
		void originXChanged(double);
		void originYChanged(double);
		void zoomChanged(double);
		void resolutionChanged(double);

public:
	void draw(QPainter& p, const QRectF& R, const QImage& image);

protected:
	virtual bool eventFilter( QObject *watched, QEvent *e );
	virtual void leaveEvent( QEvent *e );
	virtual void keyPressEvent( QKeyEvent *e );
	virtual void keyReleaseEvent( QKeyEvent *e );
	virtual void mousePressEvent( QMouseEvent *e );
	virtual void mouseMoveEvent( QMouseEvent *e );
	virtual void mouseReleaseEvent( QMouseEvent *e );
	virtual void wheelEvent(QWheelEvent *e);

	virtual void initializeGL();
	virtual void resizeGL(int w, int h);
	virtual void paintGL();

	inline void drawQuad(int width, int height, int xoff=0, int yoff=0);
	inline void drawQuadScreen(int width, int height, int xoff=0, int yoff=0);
	inline void swapToLocalMode(int w, int h);
	inline void swapToWorldMode();

protected:
	/// move tool
	QPointF m_moveOrigin;
	QPointF m_moveStart;
	QPointF m_moveEnd;
	QPointF m_moveOriginKeep;
	double m_zoom;

	/// zoom tool
	QPointF m_zoomOrigin;
	QPointF m_zoomOriginKeep;
	double m_resolution;	

	InteractionMode m_interactionMode;
	InteractionMode m_interactionModePrev;

	int	 m_interactionVertex;
	std::vector<QPointF> m_interactionEdge;
	int	m_branchIndex;

	bool m_spacePressed;
	/// interaction
	QCursor m_cursor;
	Qt::MouseButton m_dragButton;	
	QPoint	m_currPosWnd;
	QPointF  m_currPosImg;

	/// image buffer
	int m_index;
	QImage m_image;
	QImage m_record;
	//QImage recordPoint;
	bool	m_imgLoaded;
	QWidget *m_overlay;

	/// size
	QSize	m_wndSize;
	QSize	m_imgSize;
	int		m_numSpot;

	/// for dynamic focus+context
	bool	m_keyCtrl;
	bool	m_bZoomInOut;
	QTimer	*_timer;
	int		_millisec;

	int		m_frameCurr;
	int		m_frameCount;

	skeleton::SkeletonGraph m_sktGraph;
	enum SkeletonDisplay {SD_NONE=0x00, SD_EDGE=0x01, SD_VERTEX=0x02, SD_JUNCTION=0x04, SD_CONTOUR=0x08};
	bool	m_bSkeletonDisplay;
	SkeletonDisplay m_typeSkeletonDisplay;

	////////////////////////////////////////// GPU
private:
	enum {FBO1=0};

	enum {T_ORG=0,	T_GRADIENT, T_GRADIENT_S, T_PROFILE, T_PROJECT, T_NOISE, T_FIN};

	enum {P_GRADIENT, P_GRADIENT_S, P_PROFILE,  P_FIN};

	static const int NB_FBO = 1;
	static const int NB_TEX = 7;
	static const int NB_PROG = 4;

	FramebufferObject	  *_fbo[NB_FBO];
	GPUProgram            *_prog[NB_PROG];
	FloatTexture2D        *_tex[NB_TEX];

	/////////////////////////////////// parameters
	//int			m_smoothSizeUsing;
	int			m_gradSize;
	int			m_gradSmoothSize;
	double      m_widthf;
	int			m_trackingLength;

	int			m_zoomType;
	bool		m_streamlineDrawing;
	bool		m_bArrowField;
	int			m_streamlineType;

	int		m_maxStepWidth;				// ߽ĳȣ	[8,3-15:1]
	int		m_minLuminanceThre;			// صСȣ		[90,50-250,5]
	int		m_angleJunctionThre;		// жϷֲٽǶȣ  [30,0-90,1]
	int		m_angleSmoothThre;			// жϹ⻬ٽǶȣ	[150,0-180,1]

	int		m_signRadius;
	int		m_signSection;

public:
	int* gradSizePtr() {return &m_gradSize;}
	int* gradSmoothSizePtr() {return &m_gradSmoothSize;}
	double* widthfPtr() {return &m_widthf;}
	int* trackingLengthPtr() {return &m_trackingLength;}

	int* maxStepWidthPtr() {return &m_maxStepWidth;}
	int* minLuminanceThrePtr() {return &m_minLuminanceThre;}
	int* angleJunctionThrePtr() {return &m_angleJunctionThre;}
	int* angleSmoothThrePtr() {return &m_angleSmoothThre;}

	int* signRadiusPtr() {return &m_signRadius;}
	int* signSectionPtr() {return &m_signSection;}

	int* zoomTypePtr() {return &m_zoomType;}

private:
	void initGPU();
	void finalizeGPU();

	void initFBO();
	void cleanFBO();

	void initShaders();
	void setShaderParameters();
	void cleanShaders();

	void setTextureDefault(int tex_id, float val);
	//void initGaussWeightsTex(int tex_id, int smsz, float supportFactor);
	void loadTexture(int tex_id, const QImage &qimg);
	void initNoiseTex(int tex_id, const QSize &imgSize);
	//bool loadSegmentation(int tex_id, const char*filename);
	//void loadImportanceMap(int tex_id, const char*filename);

	QPointF positionZoomed(float imgx, float imgy);
	void drawStreamline();
	void drawStreamlineOne(const vector<tracking::Pot3f> &pnt, 
		const tracking::Pot3f &clr, const tracking::Pot3f &clrStart);
	void updateSignatureLabel(int px, int py, int radius, int section, bool isPolar);
	void updateProfileLabel(const float *pGrad, const float *pProf, const float *pProj,
		int proWidth, int px, int py, float dLeft, float dRight);
	void compareDifferent(const vector<tracking::Pot3f> &pnt1, const vector<tracking::Pot3f> &pnt2,
		float thres, tracking::Pot2f &diver0, tracking::Pot2f &diver1);
	void computeSkeletonGraph(skeleton::SkeletonGraph &sktGraph);
	void tracking(skeleton::SkeletonGraph &sktGraph, int w, int h, int cn, float *flow,int maxLength);
	bool trackingEdge(skeleton::SkeletonGraph & sktGraph,skeleton::Edge &pos, int w, int h, float *direction, float px, float py, int cn, float *flow);
	void centerDirection(float *tang,int w, int h, int cn, float px, float py, float *flow);
	void Record(skeleton::Edge &pos, float width, int w, int h);
	int dirJunction(float *direction, int seedx,int seedy,int w,int h,int width);
	void drawSkeletonGraph(skeleton::SkeletonGraph &sktGraph);

	void adjustPos(float & seedx, float &seedy, float anglex, float angley, int w,int h,float *flow);
	//void Record_point(float px,float py, float radius, int w, int h, int index);

	QPointF trackingInJunctionArea(const std::vector<tracking::Pot3f> &vpos, int i, 
						float seedpx, float seedpy, const QImage &img, float width);
	void drawArrowField();
};

inline void ImagePaint::drawQuad(int width, int height, int xoff/* =0 */, int yoff/* =0 */)
{
	glBegin(GL_QUADS);
	glTexCoord2i(0,0);  glVertex2i(xoff, yoff);
	glTexCoord2i(1,0);  glVertex2i(width+xoff, yoff);
	glTexCoord2i(1,1);  glVertex2i(width+xoff, height+yoff);
	glTexCoord2i(0,1);  glVertex2i(xoff, height+yoff);
	glEnd();
}

inline void ImagePaint::drawQuadScreen(int width, int height, int xoff/* =0 */, int yoff/* =0 */)
{
	glBegin(GL_QUADS);
	glTexCoord2i(0,1);  glVertex2i(xoff, yoff);
	glTexCoord2i(1,1);  glVertex2i(width+xoff, yoff);
	glTexCoord2i(1,0);  glVertex2i(width+xoff, height+yoff);
	glTexCoord2i(0,0);  glVertex2i(xoff, height+yoff);
	glEnd();
}

inline void ImagePaint::swapToLocalMode(int w, int h)
{
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0,w,0,h,-1,1);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glViewport(0,0,w,h);
	glLoadIdentity();
}

inline void ImagePaint::swapToWorldMode()
{
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
}
